//=======================================================================
// Copyright 2002 Indiana University.
// Authors: Andrew Lumsdaine, Lie-Quan Lee, Jeremy G. Siek
//
// Distributed under the Boost Software License, Version 1.0. (See
// accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)
//=======================================================================

#ifdef BOOST_GRAPH_ITERATION_MACROS_HPP

#undef BOOST_GRAPH_ITERATION_MACROS_HPP
#undef BGL_CAT
#undef BGL_FIRST
#undef BGL_LAST
#undef BGL_FORALL_VERTICES
#undef BGL_FORALL_EDGES
#undef BGL_FORALL_ADJACENT
#undef BGL_FORALL_OUTEDGES
#undef BGL_FORALL_INEDGES

#endif
