//  Copyright (c) 2001-2011 Hartmut Kaiser
//  Copyright (c) 2001-2011 Joel de Guzman
//
//  Distributed under the Boost Software License, Version 1.0. (See accompanying
//  file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)

#ifndef BOOST_SPIRIT_REPOSITORY_QI_HPP
#define BOOST_SPIRIT_REPOSITORY_QI_HPP

#if defined(_MSC_VER)
#pragma once
#endif

#include <boost/spirit/repository/home/qi/primitive.hpp>
#include <boost/spirit/repository/home/qi/directive.hpp>
#include <boost/spirit/repository/home/qi/nonterminal.hpp>
#include <boost/spirit/repository/home/qi/operator.hpp>

#endif

