#ifndef BOOST_METAPARSE_VERSION_HPP
#define BOOST_METAPARSE_VERSION_HPP

// Copyright Abel Sinkovics (abel@sinkovics.hu)  2015.
// Distributed under the Boost Software License, Version 1.0.
//    (See accompanying file LICENSE_1_0.txt or copy at
//          http://www.boost.org/LICENSE_1_0.txt)

#include <boost/predef/version_number.h>

#define BOOST_METAPARSE_VERSION BOOST_VERSION_NUMBER(1, 0, 0)

#endif

