// Copyright David Abrahams, Daniel Wallin 2003.
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

#ifndef BOOST_PARAMETER_AUX_LAMBDA_TAG_HPP
#define BOOST_PARAMETER_AUX_LAMBDA_TAG_HPP

namespace boost { namespace parameter { namespace aux {

    // Tag type passed to MPL lambda.
    struct lambda_tag;
}}} // namespace boost::parameter::aux

#endif  // include guard

